def recognize_time(time): ...
def time_to_str(t: int, units: str = "ns") -> str: ...
def time_delta_to_str(d: int) -> str: ...
def recognize_timeframe(timeframe): ...
def prec_ceil(a: float, precision: int) -> float: ...
def prec_floor(a: float, precision: int) -> float: ...
